--[[
Copyright (c) 2006-2010 Simon Ward (Greltok)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--]]

HA_timer = { accumulated = 0 }
HA_timer.__index = HA_timer

local runningTimers = {}
local frame = CreateFrame( "Frame" )

local function onUpdate( this, elapsed )
    for timer in pairs( runningTimers ) do
        timer:update( elapsed )
    end
end

function HA_timer:new( delay, callback, data, repeats )
    if not delay or delay < 0 then
        error( "delay is invalid", 2 )
    end
    if not callback then
        error( "callback is invalid", 2 )
    end
    local result = { delay = delay, callback = callback, data = data, repeats = repeats }
    setmetatable( result, self )
    return result
end

function HA_timer:run()
    if not next( runningTimers ) then
    	frame:SetScript( "OnUpdate", onUpdate )
    end
    runningTimers[self] = true
end

function HA_timer:stop()
    runningTimers[self] = nil
    if not next( runningTimers ) then
    	frame:SetScript( "OnUpdate", nil )
    end
end

function HA_timer:reset()
    self.accumulated = nil
end

function HA_timer:active()
    return runningTimers[self] and true or false
end

function HA_timer:update( elapsed )
    local newTime = self.accumulated + elapsed
    if newTime >= self.delay then
        self.callback( self.data )
        self:reset()
        if not self.repeats then
            self:stop()
        end
    else
        self.accumulated = newTime
    end
end
