--[[
Copyright (c) 2006-2010 Simon Ward (Greltok)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--]]

HA_version = {}

HA_version.__index = HA_version

function HA_version.toString( a )
    local result
    if a.bugRev and a.bugRev ~= 0 then
        result = string.format( "%d.%d.%d", ( a.majorRev or 0 ), ( a.minorRev or 0 ), a.bugRev )
    else
        result = string.format( "%d.%d", ( a.majorRev or 0 ), ( a.minorRev or 0 ) )
    end
    return result
end

function HA_version.compare( a, b )
    local result
    if ( a.majorRev or 0 ) > ( b.majorRev or 0 ) then
        result = 1
    elseif ( b.majorRev or 0 ) > ( a.majorRev or 0 ) then
        result = -1
    elseif ( a.minorRev or 0 ) > ( b.minorRev or 0 ) then
        result = 1
    elseif ( b.minorRev or 0 ) > ( a.minorRev or 0 ) then
        result = -1
    elseif ( a.bugRev or 0 ) > ( b.bugRev or 0 ) then
        result = 1
    elseif ( b.bugRev or 0 ) > ( a.bugRev or 0 ) then
        result = -1
    else
        result = 0
    end
    return result
end

HA_version.__eq = function( a, b ) return HA_version.compare( a, b ) == 0 end
HA_version.__lt = function( a, b ) return HA_version.compare( a, b ) < 0 end
HA_version.__le = function( a, b ) return HA_version.compare( a, b ) <= 0 end
HA_version.__tostring = HA_version.toString

function HA_version:new( majorRev, minorRev, bugRev )
    local result = { majorRev = majorRev, minorRev = minorRev, bugRev = bugRev }
    setmetatable( result, self )
    return result
end

function HA_version:newFromString( stringValue )
    local _, _, majorString, minorString, bugString = string.find( stringValue, "^([%d]*)%.?([%d]*)%.?([%d]*)$" )
    return self:new( tonumber( majorString ), tonumber( minorString ), tonumber( bugString ) )
end
